# Generated by Django 2.2.12 on 2020-05-11 21:55

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('fittings', '0002_add_dgm_unique_constraints'),
    ]

    operations = [
        migrations.AlterField(
            model_name='type',
            name='description',
            field=models.CharField(max_length=5000, null=True),
        ),
    ]



# Generated by Django 2.2.12 on 2020-06-05 06:44

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('groupmanagement', '0013_fix_requestlog_date_field'),
        ('fittings', '0004_add_item_category_and_group'),
    ]

    operations = [
        migrations.CreateModel(
            name='UniCategory',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(max_length=255)),
                ('color', models.TextField(max_length=20)),
                ('doctrines', models.ManyToManyField(related_name='category', to='fittings.Doctrine')),
                ('fittings', models.ManyToManyField(related_name='category', to='fittings.Fitting')),
                ('groups', models.ManyToManyField(help_text='Groups listed here will be able to access fits and doctrines listed under this category. If a category has no groups listed then it is considered an public category, accessible to anyone with permission to access permissions to the fittings module.', related_name='access_restricted_category', to='groupmanagement.AuthGroup')),
            ],
            options={
                'default_permissions': (),
            },
        ),
    ]



from allianceauth.services.hooks import get_extension_logger
from django.db import models

from .providers import esi

logger = get_extension_logger(__name__)

class ItemCategoryManager(models.Manager):
    def get_or_create(self, cat_id):
        try:
            cat = self.get(pk=cat_id)
        except:
            cat = self.create_category(cat_id)
            cat = cat[0]

        return cat

    def create_category(self, cat_id) -> tuple:
        """
        Creates ItemCategory from ESI.
        Returns a tuple containing the object and a list of related group_ids
        :param cat_id:
        :return:
        """

        c = esi.client
        cat = c.Universe.get_universe_categories_category_id(category_id=cat_id).result()
        groups = cat.pop('groups')

        obj = self.update_or_create(category_id=cat_id, defaults=cat)

        return obj[0], groups

    def create_category_with_groups(self, cat_id):
        """
        Creates category with all associated groups.
        :param cat_id:
        :return:
        """
        cat, groups = self.create_category(cat_id)

        from .models import ItemGroup
        # Check for groups missing from the category
        missing = ItemGroup.objects.check_groups(groups)

        # Create missing groups.
        for group_id in missing:
            ItemGroup.objects.create_group(group_id, cat_id)

        return cat

    @property
    def groups(self):
        """
        Returns a QueryList of the groups related to this category
        :return:
        """
        from .models import ItemGroup
        grps = ItemGroup.objects.filter(category_id=self)
        return grps


class ItemGroupManager(models.Manager):
    @property
    def types(self):
        """
        Returns a QueryList of the types related to the group.
        :return:
        """
        from .models import Type
        types = Type.objects.filter(group=self)
        return types

    def get_or_create(self, group_id):
        try:
            group = self.get(group_id=group_id)
        except:
            print("Group not found, creating.")
            group = self.create_group(group_id)

        return group

    def check_groups(self, ids: list) -> list:
        """
        Takes a list of group ids, and returns a list of the IDs that do not exist in the
        database.
        :param ids: List of ids to check
        :return: List of group ids missing.
        """
        exist = self.filter(group_id__in=ids).values_list('group_id', flat=True)
        missing = [x for x in ids if x not in exist]
        return missing

    def create_group(self, group_id, passed_id=None):
        """
        Creates group.
        :param group_id:
        :param passed_id:
        :return:
        """
        c = esi.client
        group = c.Universe.get_universe_groups_group_id(group_id=group_id).result()
        cat_id = group.pop('category_id')

        # If called from a method that has already created the category, it should send along the
        # category ID to indicate that it has been created already. We will use this to set the FK
        # and save some DB time.
        if passed_id is None:
            from .models import ItemCategory
            cat = ItemCategory.objects.get_or_create(cat_id)
            group['category'] = cat
        else:
            group['category_id'] = cat_id

        _ = group.pop('types')

        obj = self.update_or_create(group_id=group_id, defaults=group)

        return obj[0]


class TypeManager(models.Manager):
    @staticmethod
    def __get_type_id(self, type_name):
        payload = [type_name]
        c = esi.client
        ids = c.Universe.post_universe_ids(names=payload).result()
        _id = ids.get("inventory_types")[0].get("id")
        return _id

    def create_type(self, type_name):
        _id = self.__get_type_id(self, type_name)
        c = esi.client
        type_result = c.Universe.get_universe_types_type_id(type_id=_id).result()
        type_name = type_result.pop('name')
        type_result['type_name'] = type_name
        attributes = type_result.pop('dogma_attributes')
        effects = type_result.pop('dogma_effects')
        grp_id = type_result.pop('group_id')
        from .models import ItemGroup
        type_result['group'] = ItemGroup.objects.get_or_create(grp_id)

        obj = self.update_or_create(type_id=_id, defaults=type_result)

        # Handle Attributes
        from .models import DogmaAttribute
        if attributes is not None:
            DogmaAttribute.objects.bulk_attributes(attributes, obj[0].pk)

        # Handle Effects
        from .models import DogmaEffect
        if effects is not None:
            DogmaEffect.objects.bulk_effects(effects, obj[0].pk)

        return obj[0]

    def create_type_from_id(self, type_id):
        c = esi.client
        type_result = c.Universe.get_universe_types_type_id(type_id=type_id).result()
        type_name = type_result.pop('name')
        type_result['type_name'] = type_name
        attributes = type_result.pop('dogma_attributes')
        effects = type_result.pop('dogma_effects')
        grp_id = type_result.pop('group_id')
        from .models import ItemGroup
        type_result['group'] = ItemGroup.objects.get_or_create(grp_id)

        obj = self.update_or_create(type_id=type_id, defaults=type_result)

        # Handle Attributes
        from .models import DogmaAttribute
        if attributes is not None:
            DogmaAttribute.objects.bulk_attributes(attributes, obj[0].pk)

        # Handle Effects
        from .models import DogmaEffect
        if effects is not None:
            DogmaEffect.objects.bulk_effects(effects, obj[0].pk)

        return obj[0]

    # Used for the periodic update name task
    def update_type_from_id(self, type_id):
        c = esi.client
        type_result = c.Universe.get_universe_types_type_id(type_id=type_id).result()
        type_name = type_result.pop('name')
        type_result['type_name'] = type_name
        attributes = type_result.pop('dogma_attributes')
        effects = type_result.pop('dogma_effects')
        grp_id = type_result.pop('group_id')
        from .models import ItemGroup
        type_result['group'] = ItemGroup.objects.get_or_create(grp_id)

        obj = self.get(type_id=type_id)
        if obj.type_name != type_name:
            from .models import TypeHistory
            TypeHistory.objects.create(type_id=obj.type_id, type_name=obj.type_name)
            obj.type_name = type_name
            obj.save()
            
        # Handle Attributes
        from .models import DogmaAttribute
        if attributes is not None:
            DogmaAttribute.objects.bulk_attributes(attributes, obj.pk)

        # Handle Effects
        from .models import DogmaEffect
        if effects is not None:
            DogmaEffect.objects.bulk_effects(effects, obj.pk)

        return obj


class DogmaAttributeManager(models.Manager):
    def bulk_attributes(self, attributes, type_pk):
        for attribute in attributes:
            attribute['type_id'] = type_pk
            att = self.update_or_create(type_id=type_pk, attribute_id=attribute['attribute_id'], defaults=attribute)


class DogmaEffectManager(models.Manager):
    def bulk_effects(self, effects, type_pk):
        for effect in effects:
            effect['type_id'] = type_pk
            eff = self.update_or_create(type_id=type_pk, effect_id=effect['effect_id'], defaults=effect)



from allianceauth import hooks
from allianceauth.services.hooks import MenuItemHook, UrlHook

from . import urls


class FittingMenu(MenuItemHook):
    def __init__(self):
        MenuItemHook.__init__(self, 'Fittings and Doctrines',
                              'far fa-list-alt fa-fw', 'fittings:dashboard',
                              navactive=['fittings:'])

    def render(self, request):
        if request.user.has_perm('fittings.access_fittings'):
            return MenuItemHook.render(self, request)
        return ''


@hooks.register('menu_item_hook')
def register_menu():
    return FittingMenu()


@hooks.register('url_hook')
def register_url():
    return UrlHook(urls, 'fittings', '^fittings/')



from allianceauth.services.hooks import get_extension_logger
from django.contrib import messages
from django.contrib.auth.models import Group
from django.contrib.auth.decorators import login_required, permission_required
from django.utils.translation import gettext as gt
from django.db.models import Subquery, OuterRef, Count, Q, Prefetch, F
from django.shortcuts import render, redirect
from esi.decorators import token_required
from itertools import chain

from .models import Doctrine, Fitting, Type, FittingItem, Category
from .providers import esi
from .tasks import create_fit, update_fit

logger = get_extension_logger(__name__)


# Helper Function
def _get_accessible_categories(user):
    """
    Returns categories that the user has access to.
    :param user: a User object.
    :return:
    """
    groups = user.groups.all().prefetch_related('access_restricted_category')
    cats = Category.objects.none()
    for group in groups:
        cats = cats.union(group.access_restricted_category.all()\
            .annotate(groups_count=Count('groups', distinct=True))\
            .annotate(doctrines_count=Count('doctrines', distinct=True))\
            .annotate(fittings_count=Count('fittings', distinct=True))\
            .annotate(d_fittings_count=Count('doctrines__fittings', distinct=True))\
            .annotate(total_fits=F('fittings_count')+F('d_fittings_count')))
    return tuple(cats.distinct())


def _get_fits_qs(request, groups, **kwargs):
    """
    Returns a queryset representing the fits that a user can access based on their groups and/or permissions.
    Users with the manage permission will always receive all relevant fittings.
    :param request: Request object
    :param groups: Queryset representing a set of groups to be checked against.
    :param kwargs: Valid Kwargs:
       obj - An object with a MTM fittings relation to use instead of the class.
    :return: Quereyset representing the fits that the user has access to.
    """
    if 'obj' in kwargs:
        cls = kwargs['obj'].fittings
        p_fits = Fitting.objects.none()

        # If the obj is a category and is public, we need ot make sure that p_fits contains all the pure public fits for
        # the category as well as any that are also in restricted categories that the user has access to.
        if type(kwargs['obj']) is Category and len(kwargs['obj'].groups.all()) == 0:
            p_fits = cls.filter(Q(category__groups__isnull=True) | Q(doctrines__category__groups__isnull=True))\
                .exclude(Q(category__groups__isnull=False) | Q(doctrines__category__groups__isnull=False))
    else:
        cls = Fitting.objects
        p_fits = _get_public_fits()

    cls = cls.prefetch_related('category', 'doctrines__category', 'ship_type')

    if request.user.has_perm('fittings.manage'):
        fits = cls.all()
    else:
        cats = _get_accessible_categories(request.user)
        fits = cls.filter(Q(Q(category__in=cats) | Q(doctrines__category__in=cats)))
        fits = p_fits.union(fits)
    return fits.distinct()


def _get_public_fits(np=False):
    """
    Returns a list of all public fits.
    :param np: Bool flag to turn off prefetching if needed.
    :return:
    """
    if np:
        f = Fitting.objects.filter(Q(category__groups__isnull=True) | Q(doctrines__category__groups__isnull=True))\
            .exclude(Q(category__groups__isnull=False) | Q(doctrines__category__groups__isnull=False)).distinct()
    else:
        f = Fitting.objects.prefetch_related('category', 'doctrines__category', 'ship_type')\
            .filter(Q(category__groups__isnull=True) | Q(doctrines__category__groups__isnull=True))\
            .exclude(Q(category__groups__isnull=False) | Q(doctrines__category__groups__isnull=False)).distinct()
    return f


def _get_docs_qs(request, groups, **kwargs):
    """
    Returns a queryset representing the doctrines that a user can access based on their groups and/or permissions.
    Users with the manage permission will always receive all relevant doctrines.
    :param request: Request object
    :param groups: Queryset representing a set of groups to be checked against.
    :param kwargs: Valid Kwargs: obj - An object with a MTM doctrine relation to use instead of the class.
    :return: Quereyset representing the doctrines that the user has access to.
    """
    if 'obj' in kwargs:
        cls = kwargs['obj'].doctrines
    else:
        cls = Doctrine.objects

    if request.user.has_perm('fittings.manage'):
        docs = cls.prefetch_related(Prefetch('fittings', Fitting.objects.select_related('ship_type')))\
            .prefetch_related('category').all()
    else:
        docs = cls.prefetch_related('category') \
            .prefetch_related(Prefetch('fittings', Fitting.objects.select_related('ship_type'))) \
            .filter(
            Q(category__groups__in=groups))
        docs = docs.union(
            cls.prefetch_related('category')
            .prefetch_related(Prefetch('fittings', Fitting.objects.select_related('ship_type')))
            .filter(Q(category__isnull=True) | Q(category__groups__isnull=True))
        ).distinct()

    return docs


def _build_slots(fit):
    ship = fit.ship_type
    attributes = (12, 13, 14, 1137, 1367, 2056)

    t3c = ship.dogma_attributes.filter(attribute_id=1367).exists()

    attributes = ship.dogma_attributes.filter(attribute_id__in=attributes)

    slots = {'low': 0, 'med': 0, 'high': 0}
    for attribute in attributes:
        if attribute.attribute_id == 1367:
            subAttbs = (1374, 1375, 1376)
            slots['sub'] = 4
            if t3c:
                for item in FittingItem.objects.filter(fit=fit).exclude(flag='Cargo'):
                    attbs = item.type_fk.dogma_attributes.filter(attribute_id__in=subAttbs)
                    for attb in attbs:
                        if attb.attribute_id == 1374:
                            slots['high'] += int(attb.value)
                        if attb.attribute_id == 1375:
                            slots['med'] += int(attb.value)
                        if attb.attribute_id == 1376:
                            slots['low'] += int(attb.value)

        elif attribute.attribute_id == 12:
            slots['low'] += int(attribute.value)
        elif attribute.attribute_id == 13:
            slots['med'] += int(attribute.value)
        elif attribute.attribute_id == 14:
            slots['high'] += int(attribute.value)
        elif attribute.attribute_id == 1137:
            slots['rig'] = int(attribute.value)

    return slots


def _check_fit_access(request, fit_id: int) -> bool:
    fit = Fitting.objects.prefetch_related('category', 'doctrines', 'doctrines__category').get(pk=int(fit_id))
    user = request.user
    a_cats = _get_accessible_categories(user)   # Categories that the user has access to
    f_cats = fit.category.all()    # Categories attached to the fit
    for d in fit.doctrines.all():
        f_cats = chain(f_cats, d.category.all())
    pub = _get_public_fits(True)
    logger.debug(f"Checking user {user.pk} access to fit {fit_id}")
    if user.has_perm('fittings.manage'):
        logger.debug(f"User {user.pk} has manage permissions, returning True.")
        return True

    access = (fit in pub) or any(c in a_cats for c in f_cats)
    logger.debug(f"returning {access}")
    return access


# View Functions
@permission_required('fittings.access_fittings')
@login_required()
def dashboard(request):
    doc_dict = {}
    groups = request.user.groups.all()
    docs = _get_docs_qs(request, groups)

    for doc in docs:
        fits = []
        ids = []
        for fit in doc.fittings.all():
            if fit.ship_type_type_id not in ids:
                fits.append(fit)
                ids.append(fit.ship_type_type_id)
        doc_dict[doc.pk] = fits
    ctx = {'docs': docs, 'doc_dict': doc_dict}
    return render(request, 'fittings/dashboard.html', context=ctx)


@permission_required('fittings.manage')
@login_required()
def add_fit(request):
    ctx = {}
    if request.method == 'POST':
        etf_text = request.POST['eft']
        description = request.POST['description']

        create_fit.delay(etf_text, description)
        # Add success message, with note that it may take some time to see the fit on the dashboard.
        return redirect('fittings:dashboard')

    return render(request, 'fittings/add_fit.html', context=ctx)


@permission_required('fittings.manage')
@login_required()
def edit_fit(request, fit_id):
    try:
        fit = Fitting.objects.get(pk=fit_id)
        
    except Fitting.DoesNotExist:
        messages.warning(request, gt('Fit not found!'))

        return redirect('fittings:dashboard')

    if request.method == 'POST':
        etf_text = request.POST['eft']
        description = request.POST['description']

        update_fit.delay(etf_text, fit_id, description)
        # Add success message, with note that it may take some time to see the fit on the dashboard.
        return redirect('fittings:view_fit', fit_id)

    ctx = {'fit': fit}
    return render(request, 'fittings/edit_fit.html', context=ctx)


@permission_required('fittings.access_fittings')
@login_required()
def view_fit(request, fit_id):
    ctx = {}
    try:
        fit = Fitting.objects.prefetch_related('category', 'doctrines', 'doctrines__category').get(pk=fit_id)
    except Fitting.DoesNotExist:
        messages.warning(request, gt('Fit not found!'))

        return redirect('fittings:dashboard')

    # Ensure that the character should have access to the fitting.
    access = _check_fit_access(request, fit_id)

    if not access:
        messages.warning(request, gt('You do not have access to that fit.'))

        return redirect('fittings:dashboard')

    types = Type.objects.filter(type_id=OuterRef('type_id'))
    items = FittingItem.objects.filter(fit=fit).annotate(item_name=Subquery(types.values('type_name')))

    fittings = {'Cargo': [], 'FighterBay': [], 'DroneBay': []}

    for item in items:
        if item.flag == "Cargo":
            fittings['Cargo'].append(item)
        elif item.flag == "DroneBay":
            fittings['DroneBay'].append(item)
        elif item.flag == "FighterBay":
            fittings['FighterBay'].append(item)
        else:
            fittings[item.flag] = item

    ctx['doctrines'] = fit.doctrines.all()
    ctx['slots'] = _build_slots(fit)
    ctx['fit'] = fit
    ctx['fitting'] = fittings

    # Build Doctrine Category Dict
    cats = []
    ids = []
    for doc in ctx['doctrines']:
        for cat in doc.category.all():
            if cat.pk not in ids:
                cats.append(cat)
                ids.append(cat.pk)
    for cat in fit.category.all():
        if cat.pk not in ids:
            cats.append(cat)
            ids.append(cat.pk)
    del ids
    ctx['cats'] = cats

    return render(request, 'fittings/view_fit.html', context=ctx)


@permission_required('fittings.manage')
@login_required()
def add_doctrine(request):
    ctx = {}
    if request.method == 'POST':
        name = request.POST['name']
        description = request.POST['description']
        icon_url = request.POST['iconSelect']
        fitSelect = [int(fit) for fit in request.POST.getlist('fitSelect')]

        fits = Fitting.objects.filter(pk__in=fitSelect)
        d = Doctrine(name=name, description=description, icon_url=icon_url)
        d.save()
        for fitting in fits:
            d.fittings.add(fitting)
        return redirect('fittings:dashboard')

    fits = Fitting.objects.all()
    ships = Fitting.objects.order_by('ship_type').values('ship_type', 'ship_type__type_name')\
        .annotate(a=Count('ship_type'))
    ctx['fittings'] = fits
    ctx['ships'] = ships
    return render(request, 'fittings/add_doctrine.html', context=ctx)


@permission_required('fittings.access_fittings')
@login_required()
def view_doctrine(request, doctrine_id):
    ctx = {}
    try:
        doctrine = Doctrine.objects.prefetch_related('category')\
            .prefetch_related(Prefetch('fittings', Fitting.objects.select_related('ship_type')))\
            .prefetch_related('fittings__category')\
            .prefetch_related('fittings__doctrines')\
            .prefetch_related('fittings__doctrines__category').get(pk=doctrine_id)
    except Doctrine.DoesNotExist:
        messages.warning(request, gt('Doctrine not found!'))

        return redirect('fittings:dashboard')

    ctx['doctrine'] = doctrine
    ctx['d_cats'] = doctrine.category.all()
    ctx['fits'] = doctrine.fittings.all()

    # Build fit category list
    categories = dict()
    for fit in ctx['fits']:
        cats = []
        ids = []
        for cat in fit.category.all():
            if cat.pk not in ids:
                cats.append(cat)
                ids.append(cat.pk)
        for doc in fit.doctrines.all():
            for cat in doc.category.all():
                if cat.pk not in ids:
                    cats.append(cat)
                    ids.append(cat.pk)
        categories[fit.pk] = cats
        del ids
    ctx['f_cats'] = categories

    return render(request, 'fittings/view_doctrine.html', context=ctx)


@permission_required('fittings.access_fittings')
@login_required()
def view_all_fits(request):
    ctx = {}

    groups = request.user.groups.all()
    fits = _get_fits_qs(request, groups)

    ctx['fits'] = fits
    categories = dict()
    for fit in fits:
        cats = []
        ids = []
        for cat in fit.category.all():
            if cat.pk not in ids:
                cats.append(cat)
                ids.append(cat.pk)
        for doc in fit.doctrines.all():
            for cat in doc.category.all():
                if cat.pk not in ids:
                    cats.append(cat)
                    ids.append(cat.pk)
        categories[fit.pk] = cats
    ctx['cats'] = categories
    return render(request, 'fittings/view_all_fits.html', context=ctx)


@permission_required('fittings.manage')
@login_required()
def edit_doctrine(request, doctrine_id):
    ctx = {}
    try:
        doctrine = Doctrine.objects.get(pk=doctrine_id)
    except Doctrine.DoesNotExits:
        messages.warning(request, gt('Doctrine not found!'))

        return redirect('fittings:dashboard')

    if request.method == 'POST':
        name = request.POST['name']
        description = request.POST['description']
        icon_url = request.POST['iconSelect']
        fitSelect = [int(fit) for fit in request.POST.getlist('fitSelect')]
        fits = doctrine.fittings.all()

        fits = Fitting.objects.filter(pk__in=fitSelect)
        doctrine.name = name
        doctrine.description = description
        doctrine.icon_url = icon_url
        doctrine.save()
        doctrine.fittings.clear()
        for fit in fitSelect:
            doctrine.fittings.add(fit)
        return redirect('fittings:view_doctrine', doctrine_id)

    ships = Fitting.objects.order_by('ship_type').values('ship_type', 'ship_type__type_name') \
        .annotate(a=Count('ship_type'))
    ctx['ships'] = ships
    ctx['doctrine'] = doctrine
    ctx['doc_fits'] = doctrine.fittings.all()
    ctx['fits'] = Fitting.objects.exclude(pk__in=ctx['doc_fits']).all()
    return render(request, 'fittings/edit_doctrine.html', context=ctx)


@permission_required('fittings.manage')
@login_required()
def delete_doctrine(request, doctrine_id):
    try:
        doctrine = Doctrine.objects.get(pk=doctrine_id)
    except Doctrine.DoesNotExist:
        messages.warning(request, gt('Doctrine not found!'))

        return redirect('fittings:dashboard')

    doctrine.delete()

    return redirect('fittings:dashboard')


@permission_required('fittings.access_fittings')
@login_required()
def view_all_categories(request):
    ctx = {}
    if request.user.has_perm('fittings.manage'):
        cats = Category.objects\
            .all()\
            .annotate(groups_count=Count('groups', distinct=True))\
            .annotate(doctrines_count=Count('doctrines', distinct=True))\
            .annotate(fittings_count=Count('fittings', distinct=True))\
            .annotate(d_fittings_count=Count('doctrines__fittings', distinct=True))\
            .annotate(total_fits=F('fittings_count')+F('d_fittings_count'))
    else:
        cats = _get_accessible_categories(request.user)
        logger.critical(cats)
    for cat in cats:
        logger.debug(f'{cat.name}: Groups {cat.groups_count}')
    ctx['cats'] = cats
    return render(request, 'fittings/view_all_categories.html', context=ctx)


@permission_required('fittings.manage')
@login_required()
def add_category(request):
    ctx = {}
    if request.method == 'POST':
        logger.critical("POSTED")
        name = request.POST['name']
        color = request.POST['color']
        fitSelect = [int(fit) for fit in request.POST.getlist('fitSelect')]
        docSelect = [int(doc) for doc in request.POST.getlist('docSelect')]
        groupSelect = [int(grp) for grp in request.POST.getlist('groupSelect')]

        cat = Category(name=name, color=color)
        cat.save()
        for fit in fitSelect:
            cat.fittings.add(fit)
        for doc in docSelect:
            cat.doctrines.add(doc)
        for group in groupSelect:
            cat.groups.add(group)

        messages.success(request, gt("Category successfully created!"))
        return redirect('fittings:view_category', cat.pk)
    fits = Fitting.objects.all()
    docs = Doctrine.objects.all()
    groups = Group.objects.all()

    ctx = {'groups': groups, 'fits': fits, 'docs': docs}
    return render(request, 'fittings/create_category.html', ctx)


@permission_required('fittings.access_fittings')
@login_required()
def view_category(request, cat_id):
    # TODO: Check that the user has access to the category before doing anything else.
    ctx = {}
    doc_dict = {}
    try:
        cat = Category.objects\
            .prefetch_related('groups')\
            .prefetch_related('doctrines')\
            .prefetch_related('fittings')\
            .get(pk=cat_id)
        ctx['cat'] = cat
    except Exception as e:
        messages.warning(request, gt("Category not found!"))
        return redirect("fittings:dashboard")

    # Get Docs and fittings
    groups = request.user.groups.all()
    fits = _get_fits_qs(request, groups, obj=cat)
    docs = _get_docs_qs(request, groups, obj=cat)

    for doc in docs:
        fs = []
        ids = []
        for fit in doc.fittings.all():
            if fit.ship_type_type_id not in ids:
                fs.append(fit)
                ids.append(fit.ship_type_type_id)
        doc_dict[doc.pk] = fs
        del fs, ids
    ctx['docs'] = docs
    ctx['doc_dict'] = doc_dict

    categories = dict()
    for fit in fits:
        cats = []
        ids = []
        for cat in fit.category.all():
            if cat.pk not in ids:
                cats.append(cat)
                ids.append(cat.pk)
        for doc in fit.doctrines.all():
            for cat in doc.category.all():
                if cat.pk not in ids:
                    cats.append(cat)
                    ids.append(cat.pk)
        categories[fit.pk] = cats
    ctx['fits'] = fits
    ctx['cats'] = categories

    return render(request, 'fittings/view_category.html', ctx)


@permission_required('fittings.manage')
@login_required()
def edit_category(request, cat_id):
    ctx = {}
    try:
        cat = Category.objects \
            .prefetch_related('groups') \
            .prefetch_related('doctrines') \
            .prefetch_related('fittings') \
            .get(pk=cat_id)
        ctx['cat'] = cat
    except Exception as e:
        messages.warning(request, gt("Category not found!"))
        return redirect("fittings:dashboard")

    if request.method == "POST":
        name = request.POST['name']
        logger.debug(f"Updating Category {cat.name} (Now {name})")

        color = request.POST['color']
        fitSelect = [int(fit) for fit in request.POST.getlist('fitSelect')]
        docSelect = [int(doc) for doc in request.POST.getlist('docSelect')]
        groupSelect = [int(grp) for grp in request.POST.getlist('groupSelect')]

        cat.name = name
        cat.color = color
        cat.save(update_fields=['name', 'color'])
        cat.doctrines.set(docSelect)
        cat.fittings.set(fitSelect)
        cat.groups.set(groupSelect)

        messages.success(request, gt("Category successfully edited!"))
        return redirect('fittings:view_category', cat.pk)

    groups = Group.objects.all()
    doctrines = Doctrine.objects.all()
    fittings = Fitting.objects.all()

    ctx['groups'] = groups
    ctx['docs'] = doctrines
    ctx['fits'] = fittings

    return render(request, 'fittings/edit_category.html', context=ctx)


@permission_required('fittings.manage')
@login_required()
def delete_category(request, cat_id):
    ctx = {}
    try:
        cat = Category.objects.get(pk=cat_id)
    except:
        messages.warning(request, gt("Category could not be found."))
        return redirect('fittings:dashboard')
    name = cat.name
    cat.delete()

    messages.success(request, gt(f'Category "{name}" deleted successfully.'))

    return redirect('fittings:dashboard')


@permission_required('fittings.manage')
@login_required()
def delete_fit(request, fit_id):
    try:
        fit = Fitting.objects.get(pk=fit_id)
    except Doctrine.DoesNotExist:
        messages.warning(request, gt('Fit not found!'))

        return redirect('fittings:dashboard')

    fit.delete()

    return redirect('fittings:dashboard')


@permission_required('fittings.access_fittings')
@login_required()
@token_required(scopes=('esi-fittings.write_fittings.v1',))
def save_fit(request, token, fit_id):
    try:
        fit = Fitting.objects.get(pk=fit_id)
    except Fitting.DoesNotExist:
        messages.warning(request, gt('Fit not found!'))

        return redirect('fitting:dashboard')

    access = _check_fit_access(request, fit_id)
    if not access:
        messages.warning(request, gt('You do not have access to that fit.'))

        return redirect('fittings:dashboard')

    # Build POST payload
    fit_dict = {
        'description': fit.description,
        'name': fit.name,
        'ship_type_id': fit.ship_type_type_id,
        'items': []
    }
    for item in fit.items.all():
        f_item = {
            'flag': item.flag,
            'quantity': item.quantity,
            'type_id': item.type_id
        }
        fit_dict['items'].append(f_item)

    # Get client
    c = esi.client
    fit = c.Fittings\
        .post_characters_character_id_fittings(character_id=token.character_id, fitting=fit_dict,
                                               token=token.valid_access_token()).result()

    return redirect('fittings:dashboard')



# Generated by Django 2.2.12 on 2020-06-05 07:35

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('fittings', '0006_auto_20200605_0724'),
    ]

    operations = [
        migrations.AlterField(
            model_name='unicategory',
            name='doctrines',
            field=models.ManyToManyField(blank=True, help_text="All fittings in a doctrine will be treated as if they are in the doctrine's category.", related_name='category', to='fittings.Doctrine'),
        ),
        migrations.AlterField(
            model_name='unicategory',
            name='fittings',
            field=models.ManyToManyField(blank=True, help_text="Fittings only need to be tagged with a category if they are not included in any doctrines, or if they need to be labled in addition to their doctrine's category.", related_name='category', to='fittings.Fitting'),
        ),
        migrations.AlterField(
            model_name='unicategory',
            name='groups',
            field=models.ManyToManyField(blank=True, help_text='Groups listed here will be able to access fits and doctrines listed under this category. If a category has no groups listed then it is considered an public category, accessible to anyone with permission to access permissions to the fittings module.', related_name='access_restricted_category', to='groupmanagement.AuthGroup'),
        ),
    ]



# Generated by Django 2.2.12 on 2020-06-05 07:24
from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('fittings', '0005_unicategory'),
    ]

    operations = [
        migrations.AlterModelOptions(
            name='unicategory',
            options={'default_permissions': (), 'verbose_name': 'Unified Category', 'verbose_name_plural': 'Unified Categories'},
        ),
    ]



from django.contrib.auth.models import Group
from django.db import models
from django.db.models import Subquery, OuterRef
from model_utils import Choices

from .managers import TypeManager, DogmaAttributeManager, DogmaEffectManager, ItemCategoryManager, ItemGroupManager


# Category Model
class ItemCategory(models.Model):
    category_id = models.IntegerField(primary_key=True)
    name = models.CharField(max_length=250)
    published = models.BooleanField(default=True)

    objects = ItemCategoryManager()

    class Meta:
        default_permissions = ()


# Group Model
class ItemGroup(models.Model):
    group_id = models.IntegerField(primary_key=True)
    name = models.CharField(max_length=250)
    category = models.ForeignKey(ItemCategory, on_delete=models.CASCADE)
    published = models.BooleanField(default=True)

    objects = ItemGroupManager()

    class Meta:
        default_permissions = ()


# Type Model
class Type(models.Model):
    type_name = models.CharField(max_length=500)
    type_id = models.BigIntegerField(primary_key=True)
    # group_id = models.IntegerField()    # This is essentially a FK field.
    group = models.ForeignKey(ItemGroup, on_delete=models.CASCADE, null=True)
    published = models.BooleanField(default=False)
    mass = models.FloatField(null=True)
    capacity = models.FloatField(null=True)
    description = models.CharField(max_length=5000, null=True)  # Not sure of the actual max.
    volume = models.FloatField(null=True)
    packaged_volume = models.FloatField(null=True)
    portion_size = models.IntegerField(null=True)
    radius = models.FloatField(null=True)
    graphic_id = models.IntegerField(null=True)
    icon_id = models.IntegerField(null=True)
    market_group_id = models.IntegerField(null=True)

    objects = TypeManager()

    class Meta:
        default_permissions = ()


class TypeHistory(models.Model):
    type_id = models.BigIntegerField()
    type_name = models.CharField(max_length=500)

    class Meta:
        default_permissions = ()


# Dogma Attribute
class DogmaAttribute(models.Model):
    # 12 - Low Slots | 13 - Med Slots | 14 - High Slots
    # 1137 - Rig Slots | 1367 - Sub System Slots | 2056 - Service Slots
    # 182 | 183 | 184 --- Req Skill 1/2/3
    # 277 - Req. Skill 1 Lvl | 278 | 279 -- Req Skill 1/2 Lvl
    # 1374 - HiSlotModifier | 1375 - MedSlotModifier | 1376 - RigSlotModifier
    type = models.ForeignKey(Type, on_delete=models.DO_NOTHING, related_name='dogma_attributes')
    attribute_id = models.IntegerField()
    value = models.FloatField()

    objects = DogmaAttributeManager()

    class Meta:
        default_permissions = ()
        constraints = [
            models.UniqueConstraint(fields=('attribute_id', 'type'), name='unique_type_and_attribute_id')
        ]


# Dogma Effect
class DogmaEffect(models.Model):
    # 11 - Low Power | 12 - High Power | 13 - Med Power
    # 2663 - Rig Slot | 3772 - Subsystem | 6306 - Service Slot
    type = models.ForeignKey(Type, on_delete=models.DO_NOTHING, related_name='dogma_effects')
    effect_id = models.IntegerField()
    is_default = models.BooleanField()

    objects = DogmaEffectManager()

    class Meta:
        default_permissions = ()
        constraints = [
            models.UniqueConstraint(fields=('effect_id', 'type'), name='unique_type_and_effect_id')
        ]


# Fitting
class Fitting(models.Model):
    description = models.TextField(max_length=500)
    name = models.CharField(max_length=255, null=False)
    ship_type = models.ForeignKey(Type, on_delete=models.DO_NOTHING)
    ship_type_type_id = models.IntegerField()

    def __str__(self):
        return "{} ({})".format(self.ship_type.type_name, self.name)

    @property
    def eft(self):
        types = Type.objects.filter(type_id=OuterRef('type_id'))
        items = FittingItem.objects.filter(fit=self).annotate(item_name=Subquery(types.values('type_name')))

        eft = '[' + self.ship_type.type_name + ', ' + self.name + ']\n\n'

        temp = {'Cargo': [], 'FighterBay': [], 'DroneBay': []}
        
        slots = [
            {'key': 'LoSlot', 'range': 8},
            {'key': 'MedSlot', 'range': 8},
            {'key': 'HiSlot', 'range': 8},
            {'key': 'RigSlot', 'range': 3},
            {'key': 'SubSystemSlot', 'range': 4},
            {'key': 'ServiceSlot', 'range': 8}  # This is likely much higher than the actual max.
        ]
        
        for item in items:
            if item.flag == 'Cargo':
                temp['Cargo'].append(item)
            elif item.flag == 'FighterBay':
                temp['FighterBay'].append(item)
            elif item.flag == 'DroneBay':
                temp['DroneBay'].append(item)
            else:
                temp[item.flag] = item.item_name

        for slot in slots:
            is_empty = True
            for i in range(0, slot['range']):
                key = slot['key'] + str(i)
                if key in temp:
                    eft += temp[key] + '\n'
                    is_empty = False
            if not is_empty:
                eft += '\n'

        slots = [
            'FighterBay',
            'DroneBay',
            'Cargo'
        ]

        for slot in slots:
            if slot in temp:
                eft += '\n\n'
                for item in temp[slot]:
                    eft += item.item_name + ' x' + str(item.quantity) + '\n'

        eft += '\n'
        
        return eft

    class Meta:
        default_permissions = (())
        permissions = (('access_fittings', 'Can access the fittings module.'),)
        unique_together = (
            ('ship_type_type_id', 'name'),
        )


# Fitting items
class FittingItem(models.Model):
    fit = models.ForeignKey(Fitting, on_delete=models.CASCADE, related_name='items')
    _flag_enum = Choices('Cargo', 'DroneBay', 'FighterBay', 'HiSlot0', 'HiSlot1', 'HiSlot2',
                         'HiSlot3', 'HiSlot4', 'HiSlot5', 'HiSlot6', 'HiSlot7', 'Invalid',
                         'LoSlot0', 'LoSlot1', 'LoSlot2', 'LoSlot3', 'LoSlot4', 'LoSlot5',
                         'LoSlot6', 'LoSlot7', 'MedSlot0', 'MedSlot1', 'MedSlot2', 'MedSlot3',
                         'MedSlot4', 'MedSlot5', 'MedSlot6', 'MedSlot7', 'RigSlot0', 'RigSlot1',
                         'RigSlot2', 'ServiceSlot0', 'ServiceSlot1', 'ServiceSlot2', 'ServiceSlot3',
                         'ServiceSlot4', 'ServiceSlot5', 'ServiceSlot6', 'ServiceSlot7', 'SubSystemSlot0',
                         'SubSystemSlot1', 'SubSystemSlot2', 'SubSystemSlot3')
    flag = models.CharField(max_length=25, choices=_flag_enum, default='Invalid')
    quantity = models.IntegerField(default=1)
    type_fk = models.ForeignKey(Type, on_delete=models.DO_NOTHING)
    type_id = models.IntegerField()

    class Meta:
        default_permissions = (())


# Doctrine
class Doctrine(models.Model):
    name = models.CharField(max_length=255, null=False)
    icon_url = models.URLField(null=True)
    fittings = models.ManyToManyField(Fitting, related_name='doctrines')
    description = models.TextField(max_length=1000)
    created = models.DateTimeField(auto_now_add=True)
    last_updated = models.DateTimeField(null=True)

    def __str__(self):
        return f"{self.name}"

    class Meta:
        default_permissions = (())
        permissions = (("manage", "Can manage doctrines and fits."),)


# Unified Category
class Category(models.Model):
    name = models.CharField(max_length=255, null=False)
    color = models.TextField(max_length=20, default="#FFFFFF")  # Tag Color

    fittings = models.ManyToManyField(Fitting, blank=True, related_name="category",
                                      help_text="Fittings only need to be tagged with a category if they are "
                                                "not included in any doctrines, or if they need to be labled "
                                                "in addition to their doctrine's category.")
    doctrines = models.ManyToManyField(Doctrine, blank=True, related_name="category",
                                       help_text="All fittings in a doctrine will be treated as if they are in the "
                                                 "doctrine's category.")

    groups = models.ManyToManyField(Group, blank=True, related_name="access_restricted_category",
                                    help_text="Groups listed here will be able to access fits and doctrines"
                                              " listed under this category. If a category has no groups listed"
                                              " then it is considered an public category, accessible to anyone"
                                              " with permission to access permissions to the fittings module.")

    def __str__(self):
        return f"Category: {self.name}"

    class Meta:
        verbose_name = "Category"
        verbose_name_plural = "Categories"
        default_permissions = (())



# Generated by Django 2.2.8 on 2019-12-13 21:22

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('fittings', '0001_initial'),
    ]

    operations = [
        migrations.AddConstraint(
            model_name='dogmaattribute',
            constraint=models.UniqueConstraint(fields=('attribute_id', 'type'), name='unique_type_and_attribute_id'),
        ),
        migrations.AddConstraint(
            model_name='dogmaeffect',
            constraint=models.UniqueConstraint(fields=('effect_id', 'type'), name='unique_type_and_effect_id'),
        ),
    ]



# Generated by Django 2.2.12 on 2020-08-15 21:25

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('fittings', '0011_auto_20200815_2022'),
    ]

    operations = [
        migrations.CreateModel(
            name='TypeHistory',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('type_id', models.BigIntegerField()),
                ('type_name', models.CharField(max_length=500)),
            ],
            options={
                'default_permissions': (),
            },
        ),
    ]



from django.contrib import admin
from django.db.models import Count

from .forms import CategoryForm
from .models import Fitting, Category

# Register your models here.
admin.site.register(Fitting)


class CategoryAdmin(admin.ModelAdmin):
    list_select_related = True
    list_display = ('name', 'color', 'restricted', )
    list_filter = ('groups',)

    def get_queryset(self, request):
        qs = super().get_queryset(request)
        # TODO: Figure out how to get an accurate Fitting count to include doctrine fits.

        qs = qs.annotate(
            fittings_count=Count('fittings'),
            doctrine_count=Count('doctrines'),
            groups_count=Count('groups'),
        )

        return qs

    search_fields = ('name', 'color')

    def restricted(self, obj):
        return obj.groups.exists()

    restricted.boolean = True



    form = CategoryForm
    filter_horizontal = ('fittings', 'doctrines', 'groups')


admin.site.register(Category, CategoryAdmin)



# -*- coding: utf-8 -*-
from setuptools import setup, find_packages

from fittings import __version__

with open("README.md", "r") as fh:
    long_description = fh.read()

install_requires = [
    'allianceauth>=2.7.3',
    'django-bootstrap-form',
    'django-model_utils>=3.1.1',
    'django-esi>=2.0.0'
]

testing_extras = [

]

setup(
    name='fittings',
    version=__version__,
    author='Col Crunch',
    author_email='it-team@serin.space',
    description='A simple fittings and doctrine management application.',
    long_description=long_description,
    long_description_content_type="text/markdown",
    install_requires=install_requires,
    extras_require={
        'testing': testing_extras,
        ':python_version=="3.6"': ['typing'],
    },
    python_requires='~=3.6',
    license='GPLv3',
    packages=find_packages(),
    classifiers=[
        'Environment :: Web Environment',
        'Framework :: Django',
        'Framework :: Django :: 2.2',
        'Framework :: Django :: 3.1',
        'Intended Audience :: Developers',
        'Operating System :: POSIX :: Linux',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',
        'Topic :: Internet :: WWW/HTTP',
        'Topic :: Internet :: WWW/HTTP :: Dynamic Content',
    ],
    url='https://gitlab.com/colcrunch/fittings',
    zip_safe=False,
    include_package_data=True,
)


# Generated by Django 2.2.8 on 2019-12-13 03:22

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    initial = True

    dependencies = [
    ]

    operations = [
        migrations.CreateModel(
            name='Fitting',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('description', models.CharField(max_length=1000)),
                ('name', models.CharField(max_length=255)),
                ('ship_type_type_id', models.IntegerField()),
            ],
            options={
                'permissions': (('access_fittings', 'Can access the fittings module.'),),
                'default_permissions': (),
            },
        ),
        migrations.CreateModel(
            name='Type',
            fields=[
                ('type_name', models.CharField(max_length=500)),
                ('type_id', models.BigIntegerField(primary_key=True, serialize=False)),
                ('group_id', models.IntegerField()),
                ('published', models.BooleanField(default=False)),
                ('mass', models.FloatField(null=True)),
                ('capacity', models.FloatField(null=True)),
                ('description', models.CharField(max_length=5000)),
                ('volume', models.FloatField(null=True)),
                ('packaged_volume', models.FloatField(null=True)),
                ('portion_size', models.IntegerField(null=True)),
                ('radius', models.FloatField(null=True)),
                ('graphic_id', models.IntegerField(null=True)),
                ('icon_id', models.IntegerField(null=True)),
                ('market_group_id', models.IntegerField(null=True)),
            ],
            options={
                'default_permissions': (),
            },
        ),
        migrations.CreateModel(
            name='FittingItem',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('flag', models.CharField(choices=[('Cargo', 'Cargo'), ('DroneBay', 'DroneBay'), ('FighterBay', 'FighterBay'), ('HiSlot0', 'HiSlot0'), ('HiSlot1', 'HiSlot1'), ('HiSlot2', 'HiSlot2'), ('HiSlot3', 'HiSlot3'), ('HiSlot4', 'HiSlot4'), ('HiSlot5', 'HiSlot5'), ('HiSlot6', 'HiSlot6'), ('HiSlot7', 'HiSlot7'), ('Invalid', 'Invalid'), ('LoSlot0', 'LoSlot0'), ('LoSlot1', 'LoSlot1'), ('LoSlot2', 'LoSlot2'), ('LoSlot3', 'LoSlot3'), ('LoSlot4', 'LoSlot4'), ('LoSlot5', 'LoSlot5'), ('LoSlot6', 'LoSlot6'), ('LoSlot7', 'LoSlot7'), ('MedSlot0', 'MedSlot0'), ('MedSlot1', 'MedSlot1'), ('MedSlot2', 'MedSlot2'), ('MedSlot3', 'MedSlot3'), ('MedSlot4', 'MedSlot4'), ('MedSlot5', 'MedSlot5'), ('MedSlot6', 'MedSlot6'), ('MedSlot7', 'MedSlot7'), ('RigSlot0', 'RigSlot0'), ('RigSlot1', 'RigSlot1'), ('RigSlot2', 'RigSlot2'), ('ServiceSlot0', 'ServiceSlot0'), ('ServiceSlot1', 'ServiceSlot1'), ('ServiceSlot2', 'ServiceSlot2'), ('ServiceSlot3', 'ServiceSlot3'), ('ServiceSlot4', 'ServiceSlot4'), ('ServiceSlot5', 'ServiceSlot5'), ('ServiceSlot6', 'ServiceSlot6'), ('ServiceSlot7', 'ServiceSlot7'), ('SubSystemSlot0', 'SubSystemSlot0'), ('SubSystemSlot1', 'SubSystemSlot1'), ('SubSystemSlot2', 'SubSystemSlot2'), ('SubSystemSlot3', 'SubSystemSlot3')], default='Invalid', max_length=25)),
                ('quantity', models.IntegerField(default=1)),
                ('type_id', models.IntegerField()),
                ('fit', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='items', to='fittings.Fitting')),
                ('type_fk', models.ForeignKey(on_delete=django.db.models.deletion.DO_NOTHING, to='fittings.Type')),
            ],
            options={
                'default_permissions': (),
            },
        ),
        migrations.AddField(
            model_name='fitting',
            name='ship_type',
            field=models.ForeignKey(on_delete=django.db.models.deletion.DO_NOTHING, to='fittings.Type'),
        ),
        migrations.CreateModel(
            name='DogmaEffect',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('effect_id', models.IntegerField()),
                ('is_default', models.BooleanField()),
                ('type', models.ForeignKey(on_delete=django.db.models.deletion.DO_NOTHING, related_name='dogma_effects', to='fittings.Type')),
            ],
            options={
                'default_permissions': (),
            },
        ),
        migrations.CreateModel(
            name='DogmaAttribute',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('attribute_id', models.IntegerField()),
                ('value', models.FloatField()),
                ('type', models.ForeignKey(on_delete=django.db.models.deletion.DO_NOTHING, related_name='dogma_attributes', to='fittings.Type')),
            ],
            options={
                'default_permissions': (),
            },
        ),
        migrations.CreateModel(
            name='Doctrine',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(max_length=255)),
                ('icon_url', models.URLField(null=True)),
                ('description', models.CharField(max_length=1000)),
                ('created', models.DateTimeField(auto_now_add=True)),
                ('last_updated', models.DateTimeField(null=True)),
                ('fittings', models.ManyToManyField(related_name='doctrines', to='fittings.Fitting')),
            ],
            options={
                'permissions': (('manage', 'Can manage doctrines and fits.'),),
                'default_permissions': (),
            },
        ),
        migrations.AlterUniqueTogether(
            name='fitting',
            unique_together={('ship_type_type_id', 'name')},
        ),
    ]
